unit ShxFont;

interface

uses
  Winapi.Windows, System.SysUtils, System.Classes, System.Contnrs, System.Math,
  System.UITypes, Vcl.Graphics, Dialogs;

type
  //  
  TMapInfo = packed record
    WinExtX: Integer;
    WinExtY: Integer;
    VptExtX: Integer;
    VptExtY: Integer;
    VptOrgX: Integer;
    VptOrgY: Integer;
  end;

  //  
  TShxFontKind = (fkShpFont, fkUniFont, fkBigFont, fkUnknown);

  //  
  TPosition = class(TObject)
  private
    FX, FY: Double;
  public
    constructor Create(AX, AY: Double);
    property X: Double read FX;
    property Y: Double read FY;
  end;

  TPositionStack = class(TObjectStack)
  public
    function Push(APosition: TPosition): TPosition; inline; // 
    function Pop: TPosition; inline; //    
    function Peek: TPosition; inline; //    
  end;

  PCodeArray = ^TCodeArray;
  TCodeArray = array[0..15] of Boolean;

  TShxFont = class;

  // 
  TShxShape = class(TObject)
  private
    FFont: TShxFont;
    FNumber: Word;
    FBytesCount: Word;
    FName: String;
    FPenDown: Boolean;
    FGlyphScale: Double;
    FX, FY: Double;
    FXmin, FXmax, FYmin, FYmax: Double;
    FVert: Boolean;
    FGlyphHighIndex: Word;
  protected
    FGlyph: array of Byte;
    function GetGlyph(Index: Integer): Byte;
    procedure SetGlyph(Index: Integer; const Value: Byte);
    procedure SetGlyphLength(const Value: Word);
    function GetGlyphLength: Word;
    procedure DrawLine(ACanvas: TCanvas);
    procedure DrawArc(ACanvas: TCanvas; CenterX, CenterY, Radius: Double;
      CW: Boolean; StartAngle, SweepAngle: Double);
    procedure DrawOctantArc(ACanvas: TCanvas; R, DSC: Byte);
    procedure DrawFractionalArc(ACanvas: TCanvas; StartOffset, EndOffset,
      HighRadius, LowRadius, DSC: Byte);
    procedure DrawBulgeArc(ACanvas: TCanvas; dX, dY, Bulge: Byte);
    procedure DrawLenDir(ACanvas: TCanvas; const Len, Dir: Byte);
    procedure AdjustBounds(AX, AY: Double);
    procedure SpecialCodeHandler(ACanvas: TCanvas; var Index: Word); //  
  public
    constructor Create(AFont: TShxFont; ANumber, ABytesCount: Word);
    destructor Destroy; override;
    procedure GetBounds(var AXmin, AXmax, AYmin, AYmax: Double);
    procedure DrawToCanvas(ACanvas: TCanvas; var AX, AY: Double; AScale: Double);
    procedure GetIncludedCodes(ACodes: PCodeArray);
    function GetGlyphDefinition(AMaxStrLen: Byte): String;
    property Number: Word read FNumber write FNumber;
    property BytesCount: Word read FBytesCount write FBytesCount;
    property Name: String read FName write FName;
    property Glyph[Index: Integer]: Byte read GetGlyph write SetGlyph;
    property GlyphLength: Word read GetGlyphLength write SetGlyphLength;
  end;

  // 
  TShxFont = class(TObjectList)
  private
    FFontFileName: String;   //     (   LoadFromFile)
    FFontName: String;       // =   0 (    )
    FFontKind: TShxFontKind; //  
    FIsFont: Boolean;        //  " ": True  , False   
    FAbove: Byte;            //      
    FBelow: Byte;            //      
    FModes: Byte;            //   
    FUniEncoding: Byte;      //  Unicode
    FUniType: Byte;          //  
    FMaxStrLen: Byte;        // .     (   SHP)
    //   .   DrawChar(ACanvas, ACharNum, AX, AY, AMapInfo)
    FCanvas: TCanvas;        //  DrawText   nil   
    FTextHeight: Double;     //   
    FScale: Double;          //    
    FXmin, FXmax, FYmin, FYmax: Double; //    
    FArcSmoothness: Integer; //  .   
  protected
    function GetShape(Index: Integer): TShxShape;
    procedure SetShape(Index: Integer; ASample: TShxShape);
    procedure SetArcSmoothness(const Value: Integer);
    procedure SetMaxStrLen(const Value: Byte);
    procedure SetDefault;
    procedure LoadShape(const FileHandle: THandle; const ShapeNum: Word);
    procedure CalcScale;
    procedure AdjustBounds(const AX, AY: Double);
  public
    constructor Create;
    function Add(AShape: TShxShape): Integer;
    procedure Insert(Index: Integer; ASample: TShxShape);
    procedure GetBounds(var AXmin, AXmax, AYmin, AYmax: Double);
    procedure ExportToShpFile(const AFileName: String);
    function LoadFromFile(const AFileName: String): Boolean;
    procedure DrawChar(ACanvas: TCanvas; const ACharNum: Word; var AX, AY: Double; AScale: Double; AMapInfo: TMapInfo); overload;
    procedure DrawChar(ACanvas: TCanvas; const ACharNum: Word; AX, AY: Integer; AMapInfo: TMapInfo); overload;
    procedure DrawText(ACanvas: TCanvas; const AText: String; AX, AY: Integer; AMapInfo: TMapInfo);
    function GetShapeIndex(ShapeNumber: Word): Integer;
    property Shapes[Index: Integer]: TShxShape read GetShape write SetShape; default;
    property FontFileName: String read FFontFileName;
    property FontName: String read FFontName;
    property FontKind: TShxFontKind read FFontKind;
    property IsFont: Boolean read FIsFont;
    property Above: Byte read FAbove;
    property Below: Byte read FBelow;
    property Modes: Byte read FModes;
    property UniEncoding: Byte read FUniEncoding;
    property UniType: Byte read FUniType;
    property MaxStrLen: Byte read FMaxStrLen write SetMaxStrLen;
    property Scale: Double read FScale;
    property TextHeight: Double read FTextHeight write FTextHeight;
    property ArcSmoothness: Integer read FArcSmoothness write SetArcSmoothness;
  end;

  function ShxFontKindToStr(AFontKind: TShxFontKind): String;
  function SetMapInfo(DC: HDC; AMapInfo: TMapInfo): Integer;

var
  FontShx: TShxFont;

implementation

type
  TShxShapeInfo = packed record
    ShapeNumber: Word;
    GlyphBytes: Word;
  end;

  THackedShxShape = class(TShxShape);

const
  TWO_PI = Pi * 2.0;
  GLYPH_SHAPE_END: Byte = $0;
  LF: Byte = 10;
  ARC_SMOOTHNESS_MIN: Integer = 8;
  ARC_SMOOTHNESS_MAX: Integer = 20000;
  DIR_LEN: array[0..$F, 0..1] of Double = (
    ( 1.0,  0.0),
    ( 1.0,  0.5),
    ( 1.0,  1.0),
    ( 0.5,  1.0),
    ( 0.0,  1.0),
    (-0.5,  1.0),
    (-1.0,  1.0),
    (-1.0,  0.5),
    (-1.0,  0.0),
    (-1.0, -0.5),
    (-1.0, -1.0),
    (-0.5, -1.0),
    ( 0.0, -1.0),
    ( 0.5, -1.0),
    ( 1.0, -1.0),
    ( 1.0, -0.5)
  );
  MAX_STR_LEN_MIN: Byte = 8;
  MAX_STR_LEN_MAX: Byte = 128;

resourcestring
  rsShpFont = '';
  rsUniFont = '';
  rsBigFont = '';
  rsUnknown = '';
  rsFileOpenErrFmt = 'Opening File Error [%u]!';
  rsFileSeekErrFmt = 'Error Seeking File [%d, %d]!';
  rsFileReadErrFmt = 'Error Reading File [%d, %d]!';
  rsErrFontKindFmt = 'Error Font Kind [%d]!';
  rsErrBigFontIsNotSupport = 'BigFont is not support!';
  rsErrReadNumShapesFmt = 'Error Reading NumShapes [%d, %d]!';
  rsErrReadShapeInfoFmt = 'Error Reading Shape Info [%d, %d]!';
  rsErrReadShapeFmt = 'Error Reading Shape [%d, %d, %d]!';
  rsErrShapesCount = 'Error Reading Shapes [%d]!';
  rsErrShape0Corrupt = 'Error Shape 0 Corrupt [%d, %d]!';
  rsUnexpectedSpecialCodeFmt = 'Unexpected Special Code: [%d, %d, %d]!';
  rsUnexpectedGlyphEnd = 'Unexpected End of Glyph: [%d, %d]!';

var
  PosStack: TPositionStack;

function GetFontKind(AFileHeader: AnsiString): TShxFontKind;
const
  SHX_SHAPES10: AnsiString = 'AutoCAD-86 shapes 1.0';
  SHX_SHAPES11: AnsiString = 'AutoCAD-86 shapes 1.1';
  SHX_UNIFONT10: AnsiString = 'AutoCAD-86 unifont 1.0';
  SHX_BIGFONT10: AnsiString = 'AutoCAD-86 bigfont 1.0';
begin
  if (Pos(SHX_SHAPES10, AFileHeader) <> 0) or (Pos(SHX_SHAPES11, AFileHeader) <> 0) then
    Result:= fkShpFont
  else if (Pos(SHX_UNIFONT10, AFileHeader) <> 0) then
    Result:= fkUniFont
  else if (Pos(SHX_BIGFONT10, AFileHeader) <> 0) then
    Result:= fkBigFont
  else
    Result:= fkUnknown;
end;

function ShxFontKindToStr(AFontKind: TShxFontKind): String;
begin
  case AFontKind of
    fkShpFont: Result:= rsShpFont;
    fkUniFont: Result:= rsUniFont;
    fkBigFont: Result:= rsBigFont;
  else
    Result:= rsUnknown;
  end;
end;

//   "+",   DSC = (S * 16 + C),  DSC = (S * 16 + C) or $80
procedure DecodeOctant(DSC: Byte; var S, C: Byte; var CW: Boolean);
var
  tmp: Shortint;
  b: Byte;
begin
  tmp:= Shortint(DSC);
  CW:= ((tmp and $80) <> 0);
  b:= Byte(tmp);
  S:= (b and $70) shr 4;
  C:= b and $07;
end;

function EncodeOctant(DSC: Byte): String;
var
  S, C: Byte;
  CW: Boolean;
  tmp: Shortint;
  b: Byte;
begin
  tmp:= Shortint(DSC);
  CW:= ((tmp and $80) <> 0);
  b:= Byte(tmp);
  S:= (b and $70) shr 4;
  C:= b and $07;
  if CW then
    Result:= Format('-0%.1x%.1x', [S, C])
  else
    Result:= Format('0%.1x%.1x', [S, C]);
end;

//   .
// StartOctant, OctantCount = [0..7]. OctantCount = 0  8 .
procedure GetOctantArcParams(BeginX, BeginY, Radius: Double;
  StartOctant, OctantCount: Byte; Clockwise: Boolean; var CenterX, CenterY: Double;
  var StartAngle, SweepAngle: Integer);
var
  Sin, Cos: Double;
begin
  StartAngle:= StartOctant * 45;
  SinCos(DegToRad(StartAngle), Sin, Cos);
  CenterX:= BeginX - Radius * Cos;
  CenterY:= BeginY - Radius * Sin;
  if OctantCount = 0 then
    SweepAngle:= 360
  else
    SweepAngle:= 45 * OctantCount;
  if Clockwise then
    StartAngle:= 360 - StartAngle;
end;

//   .
procedure GetFractionalArcParams(BeginX, BeginY, Radius: Double;
  StartOctant, OctantCount, StartOffset, EndOffset: Byte; Clockwise: Boolean;
  var CenterX, CenterY, StartAngle, SweepAngle: Double);
var
  S: Byte;
  StartAngleInRad, StartBeta, EndBeta, EndAngle: Double;
  Sin, Cos: Extended;
begin
  StartBeta:= StartOffset * 45.0 / 256.0;
  EndBeta:= EndOffset * 45.0 / 256.0;

  if not Clockwise then //  
  begin
    S:= StartOctant;
    StartAngle:= S * 45.0 + StartBeta;
    if OctantCount <> 0 then
    begin
      if EndOffset = 0 then
        EndAngle:= (StartOctant + OctantCount) * 45.0 + EndBeta
      else
        EndAngle:= (StartOctant + OctantCount - 1) * 45.0 + EndBeta;
      SweepAngle:= EndAngle - StartAngle;
    end else
      SweepAngle:= 7.0 * 45.0 + EndBeta - StartBeta;
  end else
  begin  //  
    S:= ((8 - StartOctant) mod 8);
    StartAngle:= S * 45.0 + StartBeta;
    if OctantCount = 0 then
    begin
      EndAngle:= S * 45.0 + (OctantCount - 1) * 45.0 + EndBeta
    end else
    begin
      if EndOffset = 0 then
        EndAngle:= S * 45.0 + OctantCount * 45.0 + EndBeta
      else
        EndAngle:= S * 45.0 + (OctantCount - 1) * 45.0 + EndBeta;
    end;
    if EndAngle > 360.0 then
      EndAngle:= EndAngle - 360.0;

    if EndAngle < StartAngle then
      SweepAngle:= 360.0 + EndAngle - StartAngle
    else
      SweepAngle:= EndAngle - StartAngle;
  end;

  if SweepAngle > 360.0 then
    SweepAngle:= 360.0;

  //        
  if Clockwise then
    StartAngleInRad:= DegToRad(360.0 - StartAngle)
  else
    StartAngleInRad:= DegToRad(StartAngle);
  SinCos(StartAngleInRad, Sin, Cos);
  CenterX:= BeginX - Radius * Cos;
  CenterY:= BeginY - Radius * Sin;
end;

//   .
//   Factor = 127, .. Bulge = (2 * H / D) * 127,    Factor = 1.
procedure GetBulgeArcParams(BeginX, BeginY, dX, dY, Bulge, Factor: Double;
  var CenterX, CenterY, Radius: Double; var Clockwise: Boolean;
  var StartAngle, SweepAngle: Double);
var
  EndX, EndY, MidX, MidY: Double;
  Chord, HalfChord, H, S: Double;
begin
  Clockwise:= (Bulge < 0.0);
  //   
  EndX:= BeginX + dx;
  EndY:= BeginY + dy;
  //  
  Chord:= Sqrt((EndX - BeginX) * (EndX - BeginX) + (EndY - BeginY) * (EndY - BeginY));
  //   
  HalfChord:= Chord / 2.0;
  //  (sagitta)
  H:= Abs(Bulge) * Chord / 2.0 / Factor;
  // 
  Radius:= (HalfChord * HalfChord + H * H) / H / 2.0;
  //   -   
  MidX:= (BeginX + EndX) / 2.0;
  MidY:= (BeginY + EndY) / 2.0;
  //    
  SweepAngle:= ArcTan(Abs(Bulge / Factor)) * 4.0; //   
  if not Clockwise then // CCW -   
  begin
    if SweepAngle >= Pi then
      S:= -1.0
    else
      S:= +1.0;
  end else              // CW -   
  begin
    if SweepAngle >= Pi then
      S:= +1.0
    else
      S:= -1.0;
  end;
  //  
  CenterX:= MidX + S * Sqrt(Radius * Radius - HalfChord * HalfChord) * (BeginY - EndY) / Chord;
  CenterY:= MidY + S * Sqrt(Radius * Radius - HalfChord * HalfChord) * (EndX - BeginX) / Chord;
  // 
  StartAngle:= ArcTan2(BeginY - CenterY, BeginX - CenterX);
  if (StartAngle < 0.0) then
  begin
    if Clockwise then
      StartAngle:= Abs(StartAngle)
    else
      StartAngle:= TWO_PI + StartAngle;
  end else
  begin
    if Clockwise then
      StartAngle:= TWO_PI - StartAngle;
  end;
  StartAngle:= RadToDeg(StartAngle);
  SweepAngle:= RadToDeg(SweepAngle);
end;

function SetMapInfo(DC: HDC; AMapInfo: TMapInfo): Integer;
begin
  Result:= SetMapMode(DC, MM_ISOTROPIC);
  SetWindowExtEx(DC, AMapInfo.WinExtX, AMapInfo.WinExtY, nil);
  SetViewportExtEx(DC, AMapInfo.VptExtX, AMapInfo.VptExtY, nil);
  SetViewportOrgEx(DC, AMapInfo.VptOrgX, AMapInfo.VptOrgY, nil);
end;

{ TPosition }

constructor TPosition.Create(AX, AY: Double);
begin
  FX:= AX;
  FY:= AY;
end;

{ TPositionStack }

function TPositionStack.Peek: TPosition;
begin
  Result:= TPosition(inherited Peek);
end;

function TPositionStack.Pop: TPosition;
begin
  Result:= TPosition(inherited Pop);
end;

function TPositionStack.Push(APosition: TPosition): TPosition;
begin
  Result:= TPosition(inherited Push(APosition));
end;

{ TShxShape }

procedure TShxShape.AdjustBounds(AX, AY: Double);
begin
  if AX < FXmin then
    FXmin:= AX;
  if AX > FXmax then
    FXmax:= AX;
  if AY < FYmin then
    FYmin:= AY;
  if AY > FYmax then
    FYmax:= AY;
end;

constructor TShxShape.Create(AFont: TShxFont; ANumber, ABytesCount: Word);
begin
  FFont:= AFont;
  FNumber:= ANumber;
  FBytesCount:= ABytesCount;
end;

destructor TShxShape.Destroy;
begin
  SetLength(FGlyph, 0);
  inherited Destroy;
end;

procedure TShxShape.DrawLine(ACanvas: TCanvas);
var
  iX, iY: Integer;
begin
  //  
  AdjustBounds(FX, FY);
  // 
  if ACanvas = nil then Exit;
  iX:= Round(FX);
  iY:= Round(FY);
  if FPenDown then
  begin
    ACanvas.LineTo(iX, iY);
    ACanvas.Pixels[iX, iY]:= ACanvas.Pen.Color;
  end else
    ACanvas.MoveTo(iX, iY);
end;

// StartAngle, SweepAngle -  .
// StartAngle ..       - CW:
//  CW=False -  ;
//  CW=True -  .
procedure TShxShape.DrawArc(ACanvas: TCanvas; CenterX, CenterY, Radius: Double;
  CW: Boolean; StartAngle, SweepAngle: Double);
var
  OldPenDown: Boolean;
  EndX, EndY: Double;
  StepCnt, i: Integer;
  BeginAngle, FlareAngle, ArcLength, ArcSegmentSize, Sin, Cos: Double;
begin
  BeginAngle:= DegToRad(StartAngle); //  , .
  FlareAngle:= DegToRad(SweepAngle); //   , .
  ArcLength:= Radius * FlareAngle;   //  
  //      :      
  // ,     -      
  if CW then
  begin
    BeginAngle:= 2 * Pi - BeginAngle - FlareAngle;

    SinCos(BeginAngle, Sin, Cos);
    EndX:= CenterX + Radius * Cos;
    EndY:= CenterY + Radius * Sin;
    //       (..      )
    OldPenDown:= FPenDown;
    FPenDown:= False;
    FX:= EndX;
    FY:= EndY;
    DrawLine(ACanvas);
    FPenDown:= OldPenDown;
  end;
  //   (. .  = 3)
  ArcSegmentSize:= ArcLength / Max((FFont.ArcSmoothness * SweepAngle) / 360.0, 3.0);
  StepCnt:= Trunc(Ceil(ArcLength / ArcSegmentSize));
  for i:= 1 to StepCnt do
  begin
    SinCos(BeginAngle + FlareAngle * i / StepCnt, Sin, Cos);
    FX:= CenterX + Radius * Cos;
    FY:= CenterY + Radius * Sin;
    //  
    DrawLine(ACanvas);
  end;
  //     
{$WARNINGS OFF} //       
  if CW then
  begin
    FPenDown:= False;
    FX:= EndX;
    FY:= EndY;
    DrawLine(ACanvas);
    FPenDown:= OldPenDown;
  end;
end;
{$WARNINGS ON}

procedure TShxShape.DrawBulgeArc(ACanvas: TCanvas; dX, dY, Bulge: Byte);
var
  OfsX, OfsY: Double;
  b: Shortint;
  CenterX, CenterY, Radius: Double;
  CW: Boolean;
  StartAngle, SweepAngle: Double;
begin
  OfsX:= Shortint(dX) * FGlyphScale;
  OfsY:= Shortint(dY) * FGlyphScale;
  b:= Shortint(Bulge);
  if b <> 0 then //    -  
  begin
    GetBulgeArcParams(FX, FY, OfsX, OfsY, b, 127.0, CenterX, CenterY, Radius,
      CW, StartAngle, SweepAngle);
    DrawArc(ACanvas, CenterX, CenterY, Radius, CW, StartAngle, SweepAngle);
  end else //     -  
  begin
    FX:= FX + OfsX;
    FY:= FY + OfsY;
    DrawLine(ACanvas);
  end;
end;

procedure TShxShape.DrawFractionalArc(ACanvas: TCanvas; StartOffset, EndOffset,
  HighRadius, LowRadius, DSC: Byte);
var
  S, C: Byte;  //  ,  
  CW: Boolean; // True -   ; False -   
  CenterX, CenterY, Radius: Double;
  StartAngle, SweepAngle: Double; //  
begin
  DecodeOctant(DSC, S, C, CW);
  Radius:= ((HighRadius shl 8) + LowRadius) * FGlyphScale;
  GetFractionalArcParams(FX, FY, Radius, S, C, StartOffset, EndOffset, CW, CenterX, CenterY, StartAngle, SweepAngle);
  DrawArc(ACanvas, CenterX, CenterY, Radius, CW, StartAngle, SweepAngle);
end;

procedure TShxShape.DrawLenDir(ACanvas: TCanvas; const Len, Dir: Byte);
var
  dX, dY: Double;
begin
  dX:= DIR_LEN[Dir, 0];
  dY:= DIR_LEN[Dir, 1];
  FX:= FX + dX * Len * FGlyphScale;
  FY:= FY + dY * Len * FGlyphScale;
  DrawLine(ACanvas);
end;

procedure TShxShape.DrawOctantArc(ACanvas: TCanvas; R, DSC: Byte);
var
  S, C: Byte;  //  ,   
  CW: Boolean; // True -   ; False -   
  CenterX, CenterY, Radius: Double;
  StartAngle, SweepAngle: Integer; //  
begin
  DecodeOctant(DSC, S, C, CW);
  Radius:= R * FGlyphScale;
  GetOctantArcParams(FX, FY, Radius, S, C, CW, CenterX, CenterY, StartAngle, SweepAngle);
  DrawArc(ACanvas, CenterX, CenterY, Radius, CW, StartAngle, SweepAngle);
end;

procedure TShxShape.SpecialCodeHandler(ACanvas: TCanvas; var Index: Word);
var
  SubShapeNum: Word;
  OldPenDown: Boolean;
begin
  case FGlyph[Index] of
    $0: Inc(Index); //   
    $1: begin       //    ( )
          FPenDown:= True;
          Inc(Index);
        end;
    $2: begin       //    ( )
          FPenDown:= False;
          Inc(Index);
        end;
    $3: begin       //      
          if (Index + 1) < FGlyphHighIndex then
          begin
            if not FVert then
              FGlyphScale:= FGlyphScale / FGlyph[Index + 1];
          end else
            raise Exception.CreateFmt(rsUnexpectedSpecialCodeFmt, [Index, FGlyph[Index], FGlyphHighIndex]);
          Inc(Index, 2);
          if FVert then FVert:= False;
        end;
    $4: begin       //      
          if (Index + 1) < FGlyphHighIndex then
          begin
            if not FVert then
              FGlyphScale:= FGlyphScale * FGlyph[Index + 1];
          end else
            raise Exception.CreateFmt(rsUnexpectedSpecialCodeFmt, [Index, FGlyph[Index], FGlyphHighIndex]);
          Inc(Index, 2);
          if FVert then FVert:= False;
        end;
    $5: begin       //     
          if not FVert then
            PosStack.Push(TPosition.Create(FX, FY));
          Inc(Index);
          if FVert then FVert:= False;
        end;
    $6: begin       //     
          //    ( 13  "bold.shx" 
          //    - $6,   1   - $5)
          if not FVert and (PosStack.Count > 0) then
          begin
            FX:= PosStack.Peek.X;
            FY:= PosStack.Peek.Y;
            PosStack.Pop;
            OldPenDown:= FPenDown;
            FPenDown:= False;
            DrawLine(ACanvas);
            FPenDown:= OldPenDown;
          end;
          Inc(Index);
          if FVert then FVert:= False;
        end;
    $7: begin       //  ,      ( UNIFONT - )
          if FFont.FFontKind = fkUniFont then
          begin
            if (Index + 2) < FGlyphHighIndex then
              SubShapeNum:= (FGlyph[Index + 1] shl 8) + FGlyph[Index + 2]
            else
              raise Exception.CreateFmt(rsUnexpectedSpecialCodeFmt, [Index, FGlyph[Index], FGlyphHighIndex]);
            Inc(Index, 3);
          end else
          begin
            if (Index + 1) < FGlyphHighIndex then
              SubShapeNum:= FGlyph[Index + 1]
            else
              raise Exception.CreateFmt(rsUnexpectedSpecialCodeFmt, [Index, FGlyph[Index], FGlyphHighIndex]);
            Inc(Index, 2);
          end;
          if not FVert then
          begin
            FFont[FFont.GetShapeIndex(SubShapeNum)].DrawToCanvas(ACanvas, FX, FY, FGlyphScale);
            FPenDown:= False; //      
          end;
          if FVert then FVert:= False;
        end;
    $8: begin       //    X-Y,    
          if (Index + 2) < FGlyphHighIndex then
          begin
            if not FVert then
            begin
              FX:= FX + Shortint(FGlyph[Index + 1]) * FGlyphScale;
              FY:= FY + Shortint(FGlyph[Index + 2]) * FGlyphScale;
              DrawLine(ACanvas);
            end;
          end else
            raise Exception.CreateFmt(rsUnexpectedSpecialCodeFmt, [Index, FGlyph[Index], FGlyphHighIndex]);
          Inc(Index, 3);
          if FVert then FVert:= False;
        end;
    $9: begin       //       X-Y,   (0,0)
          Inc(Index, 2); //    dY
          while True do
          begin
            if (Index >= FGlyphHighIndex) then
              raise Exception.CreateFmt(rsUnexpectedSpecialCodeFmt, [Index - 2, FGlyph[Index - 2], FGlyphHighIndex]);
            if (FGlyph[Index - 1] = 0) and (FGlyph[Index] = 0) then // [(dX,dY) = (0,0)]?
            begin
              Inc(Index);
              Break;
            end;
            if not FVert then
            begin
              FX:= FX + Shortint(FGlyph[Index - 1]) * FGlyphScale;
              FY:= FY + Shortint(FGlyph[Index]) * FGlyphScale;
              DrawLine(ACanvas);
            end;
            Inc(Index, 2);
          end;
          if FVert then FVert:= False;
        end;
    $A: begin       //  ,     (R, DSC)
          if (Index + 2) < FGlyphHighIndex then
          begin
            if not FVert then     // ,            [,  , . ]
              DrawOctantArc(ACanvas, FGlyph[Index + 1], FGlyph[Index + 2]);
          end else
            raise Exception.CreateFmt(rsUnexpectedSpecialCodeFmt, [Index, FGlyph[Index], FGlyphHighIndex]);
          Inc(Index, 3);
          if FVert then FVert:= False;
        end;
    $B: begin       //  ,    
          if (Index + 5) < FGlyphHighIndex then
          begin
            if not FVert then         // StartOffset,   EndOffset,         HighRadius,        LowRadius,         DSC
              DrawFractionalArc(ACanvas, FGlyph[Index + 1], FGlyph[Index + 2], FGlyph[Index + 3], FGlyph[Index + 4], FGlyph[Index + 5]);
          end else
            raise Exception.CreateFmt(rsUnexpectedSpecialCodeFmt, [Index, FGlyph[Index], FGlyphHighIndex]);
          Inc(Index, 6);
          if FVert then FVert:= False;
        end;
    $C: begin       // ,   X-Y   (dX,dY,Bulge)
          if (Index + 3) < FGlyphHighIndex then
          begin
            if not FVert then    // dX,                dY,                Bulge
              DrawBulgeArc(ACanvas, FGlyph[Index + 1], FGlyph[Index + 2], FGlyph[Index + 3]);
          end else
            raise Exception.CreateFmt(rsUnexpectedSpecialCodeFmt, [Index, FGlyph[Index], FGlyphHighIndex]);
          Inc(Index, 4);
          if FVert then FVert:= False;
        end;
    $D: begin       //  ,   (dX,dY,Bulge)
          Inc(Index, 3); //    Bulge
          while True do
          begin
            if (Index > FGlyphHighIndex) then
              raise Exception.CreateFmt(rsUnexpectedSpecialCodeFmt, [Index - 3, FGlyph[Index - 3], FGlyphHighIndex]);
            if (FGlyph[Index - 2] = 0) and (FGlyph[Index - 1] = 0) then // [(dX,dY) = (0,0)]?
              Break;
            if not FVert then    // dX,                dY,                Bulge)
              DrawBulgeArc(ACanvas, FGlyph[Index - 2], FGlyph[Index - 1], FGlyph[Index]);
            Inc(Index, 3);
          end;
          if FVert then FVert:= False;
        end;
    $E: begin //       
          FVert:= True;
          Inc(Index);
        end;
    else begin //     
      if not FVert then
        DrawLenDir(ACanvas, (FGlyph[Index] and $F0) shr 4, FGlyph[Index] and $0F);
      Inc(Index);
      if FVert then FVert:= False;
    end;
  end;
end;

procedure TShxShape.DrawToCanvas(ACanvas: TCanvas; var AX, AY: Double; AScale: Double);
var
  i: Word;
begin
  FVert:= False;
  FGlyphScale:= AScale;
  FXmin:= MaxDouble;
  FXmax:= -MaxDouble;
  FYmin:= MaxDouble;
  FYmax:= -MaxDouble;
  FX:= AX;
  FY:= AY;
  //      (FX, FY)
  FPenDown:= False;
  DrawLine(ACanvas);
  FPenDown:= True;
  //  
  FGlyphHighIndex:= High(FGlyph);
  i:= 0;
  while i < FGlyphHighIndex do
  begin
    SpecialCodeHandler(ACanvas, i);
  end;
  if (i > FGlyphHighIndex) or ((i = FGlyphHighIndex) and (FGlyph[FGlyphHighIndex] <> GLYPH_SHAPE_END)) then
    raise Exception.CreateFmt(rsUnexpectedGlyphEnd, [i, FGlyphHighIndex]);
  //   = LF,   . X
  if FNumber = LF then
    FX:= 0.0 - FY;
  //     
  AX:= FX;
  AY:= FY;
  //   .  = 0,     .   
  if IsZero(FYmax - FYmin) then
    FYmin:= FFont.Above * FGlyphScale;
  //  
  FFont.AdjustBounds(FXmin, FYmin);
  FFont.AdjustBounds(FXmax, FYmax);
  FFont.GetBounds(FXmin, FXmax, FYmin, FYmax);
end;

procedure TShxShape.GetBounds(var AXmin, AXmax, AYmin, AYmax: Double);
begin
  AXmin:= FXmin;
  AXmax:= FXmax;
  AYmin:= FYmin;
  AYmax:= FYmax;
end;

function TShxShape.GetGlyph(Index: Integer): Byte;
begin
  Result:= FGlyph[Index];
end;

function TShxShape.GetGlyphDefinition(AMaxStrLen: Byte): String;
var
  i: Integer;
  n, MaxLen: Byte;
begin
  MaxLen:= EnsureRange(AMaxStrLen, MAX_STR_LEN_MIN, MAX_STR_LEN_MAX);
  Result:= Format('*%d,%d', [FNumber, GlyphLength]);
  if FName = '' then
    Result:= Result + #13#10
  else
    Result:= Format('%s,%s'#13#10, [Result, FName]);
  n:= 0;
  FGlyphHighIndex:= High(FGlyph);
  i:= Low(FGlyph);
  while i <= FGlyphHighIndex do
  begin
    case FGlyph[i] of
      $0: Result:= Format('%s0', [Result]); //   
      $1: Result:= Format('%s%d,', [Result, FGlyph[i]]); //    ( )
      $2: Result:= Format('%s%d,', [Result, FGlyph[i]]); //    ( )
      $3: begin                          //      
            if (i + 1) < FGlyphHighIndex then
              Result:= Format('%s%d,%d,', [Result, FGlyph[i], FGlyph[i + 1]])
            else
              raise Exception.CreateFmt(rsUnexpectedSpecialCodeFmt, [i, FGlyph[i], FGlyphHighIndex]);
            Inc(i);
            Inc(n);
          end;
      $4: begin                          //      
            if (i + 1) < FGlyphHighIndex then
              Result:= Format('%s%d,%d,', [Result, FGlyph[i], FGlyph[i + 1]])
            else
              raise Exception.CreateFmt(rsUnexpectedSpecialCodeFmt, [i, FGlyph[i], FGlyphHighIndex]);
            Inc(i);
            Inc(n);
          end;
      $5: Result:= Format('%s%d,', [Result, FGlyph[i]]); //     
      $6: Result:= Format('%s%d,', [Result, FGlyph[i]]); //     
      $7: begin                          //  ,      ( UNIFONT - )
            if FFont.FFontKind = fkUniFont then
            begin
              if (i + 2) < FGlyphHighIndex then
                Result:= Format('%s%d,0%.4x,', [Result, FGlyph[i], (FGlyph[i + 1] shl 8) + FGlyph[i + 2]])
              else
                raise Exception.CreateFmt(rsUnexpectedSpecialCodeFmt, [i, FGlyph[i], FGlyphHighIndex]);
              Inc(i, 2);
              Inc(n, 2);
            end else
            begin
              if (i + 1) < FGlyphHighIndex then
                Result:= Format('%s%d,%d,', [Result, FGlyph[i], FGlyph[i + 1]])
              else
                raise Exception.CreateFmt(rsUnexpectedSpecialCodeFmt, [i, FGlyph[i], FGlyphHighIndex]);
              Inc(i);
              Inc(n);
            end;
          end;
      $8: begin                          //    X-Y,    
            if (i + 2) < FGlyphHighIndex then
              Result:= Format('%s%d,(%d,%d),', [Result, FGlyph[i], Shortint(FGlyph[i + 1]), Shortint(FGlyph[i + 2])])
            else
              raise Exception.CreateFmt(rsUnexpectedSpecialCodeFmt, [i, FGlyph[i], FGlyphHighIndex]);
            Inc(i, 2);
            Inc(n, 2);
          end;
      $9: begin                          //       X-Y,   (0,0)
            Result:= Format('%s%d,', [Result, FGlyph[i]]);
            Inc(n);
            Inc(i, 2); //    dY
            while True do
            begin
              if (i >= FGlyphHighIndex) then
                raise Exception.CreateFmt(rsUnexpectedSpecialCodeFmt, [i - 2, FGlyph[i - 2], FGlyphHighIndex]);
              Result:= Format('%s(%d,%d),', [Result, Shortint(FGlyph[i - 1]), Shortint(FGlyph[i])]);
              Inc(n, 2);
              if n > MaxLen then
              begin
                Result:= Result + #13#10;
                n:= 0;
              end;
              if (FGlyph[i - 1] = 0) and (FGlyph[i] = 0) then // [(dX,dY) = (0,0)]?
                Break;
              Inc(i, 2);
            end;
          end;
      $A: begin                          //  ,    
            if (i + 2) < FGlyphHighIndex then
              Result:= Format('%s%d,(%d,%s),', [Result, FGlyph[i], FGlyph[i + 1], EncodeOctant(FGlyph[i + 2])])
            else
              raise Exception.CreateFmt(rsUnexpectedSpecialCodeFmt, [i, FGlyph[i], FGlyphHighIndex]);
            Inc(i, 2);
            Inc(n, 2);
          end;
      $B: begin                          //  ,    
            if (i + 5) < FGlyphHighIndex then
              Result:= Format('%s%d,(%d,%d,%d,%d,%s),', [Result,
                FGlyph[i], FGlyph[i + 1], FGlyph[i + 2], FGlyph[i + 3], FGlyph[i + 4], EncodeOctant(FGlyph[i + 5])])
            else
              raise Exception.CreateFmt(rsUnexpectedSpecialCodeFmt, [i, FGlyph[i], FGlyphHighIndex]);
            Inc(i, 5);
            Inc(n, 5);
          end;
      $C: begin                          // ,   X-Y  
            if (i + 3) < FGlyphHighIndex then
              Result:= Format('%s%d,(%d,%d,%d),', [Result,
                FGlyph[i], Shortint(FGlyph[i + 1]), Shortint(FGlyph[i + 2]), Shortint(FGlyph[i + 3])])
            else
              raise Exception.CreateFmt(rsUnexpectedSpecialCodeFmt, [i, FGlyph[i], FGlyphHighIndex]);
            Inc(i, 3);
            Inc(n, 3);
          end;
      $D: begin                          //  ,   X-Y  
            Result:= Format('%s%d,', [Result, FGlyph[i]]);
            Inc(i, 3); //    Bulge
            while True do
            begin
              if (i > FGlyphHighIndex) then
                raise Exception.CreateFmt(rsUnexpectedSpecialCodeFmt, [i - 3, FGlyph[i - 3], FGlyphHighIndex]);
              if (FGlyph[i - 2] = 0) and (FGlyph[i - 1] = 0) then // [(dX,dY) = (0,0)]?
              begin
                Result:= Format('%s(0,0),', [Result]);
                Dec(i);
                Break;
              end else
              begin
                Result:= Format('%s(%d,%d,%d),', [Result,
                  Shortint(FGlyph[i - 2]), Shortint(FGlyph[i - 1]), Shortint(FGlyph[i])]);
                if n > MaxLen then
                begin
                  Result:= Result + #13#10;
                  n:= 0;
                end;
              end;
              Inc(i, 3);
              Inc(n, 3);
            end;
          end;
      $E: Result:= Format('%s%d,', [Result, FGlyph[i]]); //       
    else
      Result:= Format('%s0%.2x,', [Result, FGlyph[i]]);    //     
    end;
    Inc(n);
    if n > MaxLen then
    begin
      Result:= Result + #13#10;
      n:= 0;
    end;
    Inc(i);
  end;
end;

function TShxShape.GetGlyphLength: Word;
begin
  Result:= Length(FGlyph);
end;

procedure TShxShape.GetIncludedCodes(ACodes: PCodeArray);
var
  i: Integer;
  b: Byte;
begin
  FillChar(TCodeArray(ACodes^), SizeOf(TCodeArray), 0);
  FGlyphHighIndex:= High(FGlyph);
  i:= Low(FGlyph);
  while i <= FGlyphHighIndex do
  begin
    b:= FGlyph[i];
    case b of
      $0: TCodeArray(ACodes^)[b]:= True; //   
      $1: TCodeArray(ACodes^)[b]:= True; //    ( )
      $2: TCodeArray(ACodes^)[b]:= True; //    ( )
      $3: begin                          //      
            TCodeArray(ACodes^)[b]:= True;
            Inc(i);
          end;
      $4: begin                          //      
            TCodeArray(ACodes^)[b]:= True;
            Inc(i);
          end;
      $5: TCodeArray(ACodes^)[b]:= True; //     
      $6: TCodeArray(ACodes^)[b]:= True; //     
      $7: begin                          //  ,      ( UNIFONT - )
            TCodeArray(ACodes^)[b]:= True;
            if FFont.FFontKind = fkUniFont then
              Inc(i, 2)
            else
              Inc(i);
          end;
      $8: begin                          //    X-Y,    
            TCodeArray(ACodes^)[b]:= True;
            Inc(i, 2);
          end;
      $9: begin                          //       X-Y,   (0,0)
            TCodeArray(ACodes^)[b]:= True;
            Inc(i, 2); //    dY
            while True do
            begin
              if (i >= FGlyphHighIndex) then
                raise Exception.CreateFmt(rsUnexpectedSpecialCodeFmt, [i - 2, FGlyph[i - 2], FGlyphHighIndex]);
              if (FGlyph[i - 1] = 0) and (FGlyph[i] = 0) then // [(dX,dY) = (0,0)]?
                Break;
              Inc(i, 2);
            end;
          end;
      $A: begin                          //  ,    
            TCodeArray(ACodes^)[b]:= True;
            Inc(i, 2);
          end;
      $B: begin                          //  ,    
            TCodeArray(ACodes^)[b]:= True;
            Inc(i, 5);
          end;
      $C: begin                          // ,   X-Y  
            TCodeArray(ACodes^)[b]:= True;
            Inc(i, 3);
          end;
      $D: begin                          //  ,  
            TCodeArray(ACodes^)[b]:= True;
            Inc(i, 3); //    Bulge
            while True do
            begin
              if (i > FGlyphHighIndex) then
                raise Exception.CreateFmt(rsUnexpectedSpecialCodeFmt, [i - 3, FGlyph[i - 3], FGlyphHighIndex]);
              if (FGlyph[i - 2] = 0) and (FGlyph[i - 1] = 0) then // [(dX,dY) = (0,0)]?
              begin
                Dec(i);
                Break;
              end;
              Inc(i, 3);
            end;
          end;
      $E: TCodeArray(ACodes^)[b]:= True; //       
    else
      TCodeArray(ACodes^)[$F]:= True;    //     
    end;
    Inc(i);
  end;
end;

procedure TShxShape.SetGlyph(Index: Integer; const Value: Byte);
begin
  FGlyph[Index]:= Value;
end;

procedure TShxShape.SetGlyphLength(const Value: Word);
begin
  SetLength(FGlyph, Value);
end;

{ TShxShapes }

function TShxFont.Add(AShape: TShxShape): Integer;
begin
  Result:= inherited Add(AShape);
end;

procedure TShxFont.AdjustBounds(const AX, AY: Double);
begin
  if AX < FXmin then
    FXmin:= AX;
  if AX > FXmax then
    FXmax:= AX;
  if AY < FYmin then
    FYmin:= AY;
  if AY > FYmax then
    FYmax:= AY;
end;

procedure TShxFont.CalcScale;
var
  n: Integer;
  m: Double;
begin
  if FCanvas <> nil then
  begin
    m:= FTextHeight / FAbove;
    n:= GetDeviceCaps(FCanvas.Handle, LOGPIXELSX);
    FScale:= m * n / 25.4;
  end else
    FScale:= 1.0;
end;

constructor TShxFont.Create;
begin
  inherited Create;
  FCanvas:= nil;
  SetDefault;
  FMaxStrLen:= 10;
  FFontKind:= fkUnknown;
  FTextHeight:= 50.0;
  FArcSmoothness:= 8;
  CalcScale;
end;

procedure TShxFont.DrawChar(ACanvas: TCanvas; const ACharNum: Word; var AX,
  AY: Double; AScale: Double; AMapInfo: TMapInfo);
var
  n: Integer;
begin
  n:= GetShapeIndex(ACharNum);
  if n > -1 then //   
  begin
    if FCanvas <> nil then
      SetMapInfo(FCanvas.Handle, AMapInfo);
    Shapes[n].DrawToCanvas(FCanvas, AX, AY, AScale);
  end;
end;

procedure TShxFont.DrawChar(ACanvas: TCanvas; const ACharNum: Word; AX,
  AY: Integer; AMapInfo: TMapInfo);
var
  X, Y: Double;
begin
  FCanvas:= ACanvas;
  try
    if FCanvas <> nil then
      SetMapInfo(FCanvas.Handle, AMapInfo);
    CalcScale;
    FXmin:= MaxDouble;
    FXmax:= -MaxDouble;
    FYmin:= MaxDouble;
    FYmax:= -MaxDouble;
    X:= AX;
    Y:= AY;
    if FCanvas <> nil then
      FCanvas.Pen.Width:= 1;
    DrawChar(ACanvas, ACharNum, X, Y, FScale, AMapInfo);
  finally
    FCanvas:= nil;
  end;
end;

procedure TShxFont.DrawText(ACanvas: TCanvas; const AText: String; AX, AY: Integer; AMapInfo: TMapInfo);
var
  i: Integer;
  X, Y: Double;
begin
  FCanvas:= ACanvas;
  try
    if FCanvas <> nil then
      SetMapInfo(FCanvas.Handle, AMapInfo);
    CalcScale;
    FXmin:= MaxDouble;
    FXmax:= -MaxDouble;
    FYmin:= MaxDouble;
    FYmax:= -MaxDouble;
    X:= AX;
    Y:= AY;
    if FCanvas <> nil then
      FCanvas.Pen.Width:= 1;
    for i:= 1 to Length(AText) do
      DrawChar(ACanvas, Ord(AText[i]), X, Y, FScale, AMapInfo);
  finally
    FCanvas:= nil;
  end;
end;

procedure TShxFont.ExportToShpFile(const AFileName: String);
var
  Lst: TStringList;
  i: Integer;
begin
  if Count < 1 then Exit;
  Lst:= TStringList.Create;
  try
    if FFontKind = fkShpFont then
    begin
      Lst.Add(Format('*0,4,%s', [FFontName]));
      Lst.Add(Format('%d,%d,%d,0', [FAbove, FBelow, FModes]));
    end else // fkUniFont
    begin
      Lst.Add(Format('*UNIFONT,6,%s', [FFontName]));
      Lst.Add(Format('%d,%d,%d,%d,%d,0', [FAbove, FBelow, FModes, FUniEncoding, FUniType]));
    end;
    for i:= 0 to Count - 1 do
      Lst.Add(Shapes[i].GetGlyphDefinition(FMaxStrLen));
    Lst.Add('');
    Lst.SaveToFile(AFileName);
  finally
    Lst.Free;
  end;
end;

procedure TShxFont.GetBounds(var AXmin, AXmax, AYmin, AYmax: Double);
begin
  AXmin:= FXmin;
  AXmax:= FXmax;
  AYmin:= FYmin;
  AYmax:= FYmax;
end;

function TShxFont.GetShape(Index: Integer): TShxShape;
begin
  Result:= TShxShape(inherited Items[Index]);
end;

function TShxFont.GetShapeIndex(ShapeNumber: Word): Integer;
var
  First, Last, Mid: Integer;
begin
  Result:= -1;
  if Count = 0 then Exit;
  First:= 0;
  Last:= Count - 1;
  //  
  while (First < Last) do
  begin
    Mid:= First + (Last - First) div 2;
    if Shapes[Mid].Number >= ShapeNumber then
      Last:= Mid
    else
      First:= Mid + 1;
  end;
  if Shapes[Last].Number = ShapeNumber then
    Result:= Last;
end;

procedure TShxFont.SetArcSmoothness(const Value: Integer);
begin
  FArcSmoothness:= EnsureRange(Value, ARC_SMOOTHNESS_MIN, ARC_SMOOTHNESS_MAX);
end;

procedure TShxFont.SetDefault;
begin
  FIsFont:= False;
  FAbove:= 10;
  FBelow:= 3;
  FModes:= 0;
  FUniEncoding:= 2;
  FUniType:= 0;
  FFontName:= '';
  FFontFileName:= '';
end;

procedure TShxFont.SetMaxStrLen(const Value: Byte);
begin
  FMaxStrLen:= EnsureRange(Value, MAX_STR_LEN_MIN, MAX_STR_LEN_MAX);
end;

procedure TShxFont.SetShape(Index: Integer; ASample: TShxShape);
begin
  inherited Items[Index]:= ASample;
end;

procedure TShxFont.Insert(Index: Integer; ASample: TShxShape);
begin
  inherited Insert(Index, ASample);
end;

procedure TShxFont.LoadShape(const FileHandle: THandle; const ShapeNum: Word);
var
  ReadCount: Integer;
  GlyphLen, GlyphOfs: Word;
  Buffer: PAnsiChar;
  S: AnsiString;
begin
  Buffer:= System.AllocMem(Shapes[ShapeNum].BytesCount + 1);
  try
    ReadCount:= FileRead(FileHandle, Buffer^, Shapes[ShapeNum].BytesCount);
    if ReadCount <> Shapes[ShapeNum].BytesCount then
      raise Exception.CreateFmt(rsErrReadShapeFmt, [ShapeNum, ReadCount, Shapes[ShapeNum].BytesCount]);
    S:= AnsiString(Buffer);
    Shapes[ShapeNum].Name:= String(S);
    //  
    GlyphOfs:= Length(Shapes[ShapeNum].Name) + 1;
    GlyphLen:= Shapes[ShapeNum].BytesCount - GlyphOfs;
    Shapes[ShapeNum].GlyphLength:= GlyphLen;
    System.Move(Buffer[GlyphOfs], THackedShxShape(Shapes[ShapeNum]).FGlyph[0], GlyphLen);
  finally
    FreeMem(Buffer);
  end;
end;

function TShxFont.LoadFromFile(const AFileName: String): Boolean;
const
  SHX_FILE_SIGNATURE_SIZE: Word = 32;     // ,     
  SHX_SHAPES_NUM_SHAPES_OFS: Word = $1C;  //       
  SHX_UNIFONT_NUM_SHAPES_OFS: Word = $19; //       
  SHX_SHAPES_SHAPE0_GLYPH_LENGTH: Word = $03;   //       0
  SHX_UNIFNT_SHAPE0_GLYPH_LENGTH: Word = $05;   //       0
var
  FileHandle: THandle;
  ReadCount, NumShapesOfs: Integer;
  i, NumShapes, Shape0GlyphLen: Word;
  ShapeInfo: TShxShapeInfo;
  Buffer: PAnsiChar;
  S: AnsiString;
begin
  Result:= False;
  Clear;
  FileHandle:= FileOpen(AFileName, fmOpenRead);
  if FileHandle = INVALID_HANDLE_VALUE then
    raise Exception.CreateFmt(rsFileOpenErrFmt, [GetLastError]);
  try
    if FileSeek(FileHandle, 0, FILE_BEGIN) < 0 then
      raise Exception.CreateFmt(rsFileSeekErrFmt, [0, FILE_BEGIN]);
    //    -  
    Buffer:= AllocMem(SHX_FILE_SIGNATURE_SIZE + 1);
    try
      ReadCount:= FileRead(FileHandle, Buffer^, SHX_FILE_SIGNATURE_SIZE);
      if ReadCount <> SHX_FILE_SIGNATURE_SIZE then
        raise Exception.CreateFmt(rsFileReadErrFmt, [ReadCount, SHX_FILE_SIGNATURE_SIZE]);
      S:= AnsiString(Buffer);
    finally
      FreeMem(Buffer);
    end;
    FFontKind:= GetFontKind(S);
    if FFontKind = fkUnknown then
      raise Exception.CreateFmt(rsErrFontKindFmt, [Integer(FFontKind)]);
    if FFontKind = fkBigFont then
      raise Exception.Create(rsErrBigFontIsNotSupport);
    //  
    if FFontKind = fkShpFont then
      NumShapesOfs:= SHX_SHAPES_NUM_SHAPES_OFS
    else
      NumShapesOfs:= SHX_UNIFONT_NUM_SHAPES_OFS;
    if FileSeek(FileHandle, NumShapesOfs, FILE_BEGIN) < 0 then
      raise Exception.CreateFmt(rsFileSeekErrFmt, [NumShapesOfs, FILE_BEGIN]);
    ReadCount:= FileRead(FileHandle, NumShapes, SizeOf(NumShapes));
    if ReadCount <> SizeOf(NumShapes) then
      raise Exception.CreateFmt(rsErrReadNumShapesFmt, [ReadCount, SizeOf(NumShapes)]);
    //  SHAPE-  : [ ][  (,  )]
    if FFontKind = fkShpFont then
    begin
      for i:= 0 to NumShapes - 1 do
      begin
        ReadCount:= FileRead(FileHandle, ShapeInfo, SizeOf(TShxShapeInfo));
        if ReadCount <> SizeOf(TShxShapeInfo) then
          raise Exception.CreateFmt(rsErrReadShapeInfoFmt, [ReadCount, SizeOf(TShxShapeInfo)]);
        Add(TShxShape.Create(Self, ShapeInfo.ShapeNumber, ShapeInfo.GlyphBytes));
      end;
    end;
    //    
    for i:= 0 to NumShapes - 1 do
    begin
      case FFontKind of
        fkShpFont: LoadShape(FileHandle, i);
        fkUniFont: begin
          ReadCount:= FileRead(FileHandle, ShapeInfo, SizeOf(TShxShapeInfo));
          if ReadCount <> SizeOf(TShxShapeInfo) then
            raise Exception.CreateFmt(rsErrReadShapeInfoFmt, [ReadCount, SizeOf(TShxShapeInfo)]);
          Add(TShxShape.Create(Self, ShapeInfo.ShapeNumber, ShapeInfo.GlyphBytes));
          LoadShape(FileHandle, i);
        end;
      end;
    end;
    if Count < 1 then
      raise Exception.CreateFmt(rsErrShapesCount, [Count]);
    //    
    if Shapes[0].Number = 0 then //   
    begin
      FIsFont:= True;
      //      0
      if FFontKind = fkShpFont then
        Shape0GlyphLen:= SHX_SHAPES_SHAPE0_GLYPH_LENGTH
      else
        Shape0GlyphLen:= SHX_UNIFNT_SHAPE0_GLYPH_LENGTH;
      if Shapes[0].GlyphLength < Shape0GlyphLen then
        raise Exception.CreateFmt(rsErrShape0Corrupt, [Shapes[0].GlyphLength, Shape0GlyphLen]);
      //    fkShpFont   fkUniFont
      FAbove:= Shapes[0].Glyph[0];
      FBelow:= Shapes[0].Glyph[1];
      FModes:= Shapes[0].Glyph[2];
      //   fkUniFont
      if FFontKind = fkUniFont then
      begin
        FUniEncoding:= Shapes[0].Glyph[3];
        FUniType:= Shapes[0].Glyph[4];
      end;
      if Shapes[0].Name <> '' then
        FFontName:= Shapes[0].Name
      else
        FFontName:= ChangeFileExt(FFontFileName, '');
      //      0,   
      Delete(0);
    end else //   
      SetDefault;
    FFontFileName:= AFileName;
    Result:= True;
  finally
    FileClose(FileHandle);
  end;
end;

initialization
  PosStack:= TPositionStack.Create;
  FontShx:= TShxFont.Create;

finalization
  PosStack.Free;
  FontShx.Free;

end.
 